import re

### MAPPINGS ###

_direction_number_to_direction_symbol = {
    0: '',
    1: '+',
    -1: '-',
}

_accidental_abbreviation_to_name = {
    'ss': 'double sharp',
    'tqs': 'three-quarters sharp',
    's': 'sharp',
    'qs': 'quarter sharp',
    '': 'natural',
    'qf': 'quarter flat',
    'f': 'flat',
    'tqf': 'three-quarters flat',
    'ff': 'double flat',
}

_accidental_abbreviation_to_semitones = {
    'ff': -2,
    'tqf': -1.5,
    'f': -1,
    'qf': -0.5,
    '': 0,
    'qs': 0.5,
    's': 1,
    'tqs': 1.5,
    'ss': 2,
}

_accidental_abbreviation_to_symbol = {
    'ff': 'bb',
    'tqf': 'b~',
    'f': 'b',
    'qf': '~',
    '': '',
    'qs': '+',
    's': '#',
    'tqs': '#+',
    'ss': '##',
}

_accidental_name_to_abbreviation = {
    'double sharp': 'ss',
    'three-quarters sharp': 'tqs',
    'sharp': 's',
    'quarter sharp': 'qs',
    'natural': '',
    'quarter flat': 'qf',
    'flat': 'f',
    'three-quarters flat': 'tqf',
    'double flat': 'ff',
}

_accidental_semitones_to_abbreviation = {
    -2.0: 'ff',
    -1.5: 'tqf',
    -1.0: 'f',
    -0.5: 'qf',
    0.0: '',
    0.5: 'qs',
    1.0: 's',
    1.5: 'tqs',
    2.0: 'ss',
}

_symbolic_accidental_to_abbreviation = {
    'bb': 'ff',
    'b~': 'tqf',
    'b': 'f',
    '~': 'qf',
    '': '',
    '!': '!',
    '+': 'qs',
    '#': 's',
    '#+': 'tqs',
    '##': 'ss',
}

_symbolic_accidental_to_semitones = {
    'bb': -2,
    'b~': -1.5,
    'b': -1,
    '~': -0.5,
    '': 0,
    '+': 0.5,
    '#': 1,
    '#+': 1.5,
    '##': 2,
    'ff': -2,
    'tqf': 1.5,
    'f': -1,
    'qf': -0.5,
    'qs': 0.5,
    's': 1,
    'tqs': 1.5,
    'ss': 2,
}

_diatonic_pc_name_to_diatonic_pc_number = {
    'c': 0,
    'd': 1,
    'e': 2,
    'f': 3,
    'g': 4,
    'a': 5,
    'b': 6,
}

_diatonic_pc_name_to_pitch_class_number = {
    'c': 0,
    'd': 2,
    'e': 4,
    'f': 5,
    'g': 7,
    'a': 9,
    'b': 11,
}

_diatonic_pc_number_to_diatonic_pc_name = {
    0: 'c',
    1: 'd',
    2: 'e',
    3: 'f',
    4: 'g',
    5: 'a',
    6: 'b',
}

_diatonic_pc_number_to_pitch_class_number = {
    0: 0,
    1: 2,
    2: 4,
    3: 5,
    4: 7,
    5: 9,
    6: 11,
}

_pitch_class_number_to_diatonic_pc_number = {
    0: 0,
    2: 1,
    4: 2,
    5: 3,
    7: 4,
    9: 5,
    11: 6,
}

_pitch_class_number_to_pitch_class_name = {
    0.0: 'c',
    0.5: 'cqs',
    1.0: 'cs',
    1.5: 'dqf',
    2.0: 'd',
    2.5: 'dqs',
    3.0: 'ef',
    3.5: 'eqf',
    4.0: 'e',
    4.5: 'eqs',
    5.0: 'f',
    5.5: 'fqs',
    6.0: 'fs',
    6.5: 'gqf',
    7.0: 'g',
    7.5: 'gqs',
    8.0: 'af',
    8.5: 'aqf',
    9.0: 'a',
    9.5: 'aqs',
    10.0: 'bf',
    10.5: 'bqf',
    11.0: 'b',
    11.5: 'bqs',
}

_pitch_class_number_to_pitch_class_name_with_flats = {
    0.0: 'c',
    0.5: 'dtqf',
    1.0: 'df',
    1.5: 'dqf',
    2.0: 'd',
    2.5: 'etqf',
    3.0: 'ef',
    3.5: 'eqf',
    4.0: 'e',
    4.5: 'fqf',
    5.0: 'f',
    5.5: 'gtqf',
    6.0: 'gf',
    6.5: 'gqf',
    7.0: 'g',
    7.5: 'atqf',
    8.0: 'af',
    8.5: 'aqf',
    9.0: 'a',
    9.5: 'btqf',
    10.0: 'bf',
    10.5: 'bqf',
    11.0: 'b',
    11.5: 'cqf',
}

_pitch_class_number_to_pitch_class_name_with_sharps = {
    0.0: 'c',
    0.5: 'cqs',
    1.0: 'cs',
    1.5: 'ctqs',
    2.0: 'd',
    2.5: 'dqs',
    3.0: 'ds',
    3.5: 'dtqs',
    4.0: 'e',
    4.5: 'eqs',
    5.0: 'f',
    5.5: 'fqs',
    6.0: 'fs',
    6.5: 'ftqs',
    7.0: 'g',
    7.5: 'gqs',
    8.0: 'gs',
    8.5: 'gtqs',
    9.0: 'a',
    9.5: 'aqs',
    10.0: 'as',
    10.5: 'atqs',
    11.0: 'b',
    11.5: 'bqs',
}

_diatonic_number_and_quality_to_semitones = {
    1: {'d': -1, 'P': 0, 'A': 1},
    2: {'d': 0, 'm': 1, 'M': 2, 'A': 3},
    3: {'d': 2, 'm': 3, 'M': 4, 'A': 5},
    4: {'d': 4, 'P': 5, 'A': 6},
    5: {'d': 6, 'P': 7, 'A': 8},
    6: {'d': 7, 'm': 8, 'M': 9, 'A': 10},
    7: {'d': 9, 'm': 10, 'M': 11, 'A': 12},
    8: {'d': 11, 'P': 12, 'A': 13},
}

_semitones_to_quality_and_diatonic_number = {
    0: ('P', 1),
    1: ('m', 2),
    2: ('M', 2),
    3: ('m', 3),
    4: ('M', 3),
    5: ('P', 4),
    6: ('d', 5),
    7: ('P', 5),
    8: ('m', 6),
    9: ('M', 6),
    10: ('m', 7),
    11: ('M', 7),
    12: ('P', 8),
}

_quality_abbreviation_to_quality_string = {
    'M': 'major',
    'm': 'minor',
    'P': 'perfect',
    'aug': 'augmented',
    'dim': 'diminished',
    'A': 'augmented',
    'd': 'diminished',
    }

_quality_string_to_quality_abbreviation = {
    'major': 'M',
    'minor': 'm',
    'perfect': 'P',
    'augmented': 'A',
    'diminished': 'd',
    }

_semitones_to_quality_string_and_number = {
    0: ('perfect', 1),
    1: ('minor', 2),
    2: ('major', 2),
    3: ('minor', 3),
    4: ('major', 3),
    5: ('perfect', 4),
    6: ('diminished', 5),
    7: ('perfect', 5),
    8: ('minor', 6),
    9: ('major', 6),
    10: ('minor', 7),
    11: ('major', 7),
    }

_start_punctuation_to_inclusivity_string = {
    '[': 'inclusive',
    '(': 'exclusive',
}

_stop_punctuation_to_inclusivity_string = {
    ']': 'inclusive',
    ')': 'exclusive',
}

### REGEX ATOMS ###

_integer_regex_atom = '-?\d+'

_alphabetic_accidental_regex_atom = (
    '(?P<alphabetic_accidental>'
    '[s]*(qs)?'
    '|[f]*(qf)?'
    '|t?q?[fs]'
    '|'
    ')'
)

_symbolic_accidental_regex_atom = (
    '(?P<symbolic_accidental>'
    '[#]+[+]?'
    '|[b]+[~]?'
    '|[+]'
    '|[~]'
    '|'
    ')'
)

_octave_number_regex_atom = (
    '(?P<octave_number>{}|)'.format(_integer_regex_atom)
)

_octave_tick_regex_atom = (
    '(?P<octave_tick>'
    ',+'
    "|'+"
    '|'
    ')'
)

_diatonic_pc_name_regex_atom = (
    '(?P<diatonic_pc_name>'
    '[A-Ga-g]'
    ')'
)

### REGEX BODIES ###

_comprehensive_accidental_regex_body = (
    '(?P<comprehensive_accidental>{}|{})'
).format(
    _alphabetic_accidental_regex_atom,
    _symbolic_accidental_regex_atom,
)

_comprehensive_octave_regex_body = (
    '(?P<comprehensive_octave>{}|{})'
).format(
    _octave_number_regex_atom,
    _octave_tick_regex_atom,
)

_comprehensive_pitch_class_name_regex_body = (
    '(?P<comprehensive_pitch_class_name>{}{})'
).format(
    _diatonic_pc_name_regex_atom,
    _comprehensive_accidental_regex_body,
)

_comprehensive_pitch_name_regex_body = (
    '(?P<comprehensive_pitch_name>{}{}{})'
).format(
    _diatonic_pc_name_regex_atom,
    _comprehensive_accidental_regex_body,
    _comprehensive_octave_regex_body,
)

_pitch_class_name_regex_body = (
    '(?P<pitch_class_name>{}{})'
).format(
    _diatonic_pc_name_regex_atom,
    _alphabetic_accidental_regex_atom,
)

_pitch_class_octave_number_regex_body = (
    '(?P<pitch_class_octave_number>{}{}{})'
).format(
    _diatonic_pc_name_regex_atom,
    _comprehensive_accidental_regex_body,
    _octave_number_regex_atom,
)

_pitch_name_regex_body = (
    '(?P<pitch_name>{}{}{})'
).format(
    _diatonic_pc_name_regex_atom,
    _alphabetic_accidental_regex_atom,
    _octave_tick_regex_atom,
)

_range_string_regex_body = '''
    (?P<open_bracket>
        [\[(]       # open bracket or open parenthesis
    )
    (?P<start_pitch>
        {}|{}|(?P<start_pitch_number>-?\d+) # start pitch
    )
    ,               # comma
    [ ]*            # any amount of whitespace
    (?P<stop_pitch>
        {}|{}|(?P<stop_pitch_number>-?\d+) # stop pitch
    )
    (?P<close_bracket>
        [\])]       # close bracket or close parenthesis
    )
    '''.format(
    _pitch_class_octave_number_regex_body.replace('<', '<us_start_'),
    _pitch_name_regex_body.replace('<', '<ly_start_'),
    _pitch_class_octave_number_regex_body.replace('<', '<us_stop_'),
    _pitch_name_regex_body.replace('<', '<ly_stop_'),
)

_interval_name_abbreviation_regex_body = '''
    (?P<direction>[+,-]?)  # one plus, one minus, or neither
    (?P<quality>           # exactly one quality abbreviation
        M|                 # major
        m|                 # minor
        P|                 # perfect
        aug|               # augmented
        A+|                # (possibly) multi-augmented
        dim|               # dimished
        d+                 # (possibly) multi-diminished
    )
    (?P<quartertone>[+~]?) # followed by an optional quartertone inflection
    (?P<number>\d+)        # followed by one or more digits
    '''

### REGEX PATTERNS ###

_alphabetic_accidental_regex = re.compile(
    '^{}$'.format(_alphabetic_accidental_regex_atom),
    re.VERBOSE,
)

_symbolic_accidental_regex = re.compile(
    '^{}$'.format(_symbolic_accidental_regex_atom),
    re.VERBOSE,
)

_comprehensive_accidental_regex = re.compile(
    '^{}$'.format(_comprehensive_accidental_regex_body),
    re.VERBOSE,
)

_octave_tick_regex = re.compile(
    '^{}$'.format(_octave_tick_regex_atom),
    re.VERBOSE,
)

_octave_number_regex = re.compile(
    '^{}$'.format(_octave_number_regex_atom),
    re.VERBOSE,
)

_diatonic_pc_name_regex = re.compile(
    '^{}$'.format(_diatonic_pc_name_regex_atom),
    re.VERBOSE,
)

_comprehensive_accidental_regex = re.compile(
    '^{}$'.format(_comprehensive_accidental_regex_body),
    re.VERBOSE,
)

_comprehensive_octave_regex = re.compile(
    '^{}$'.format(_comprehensive_octave_regex_body),
    re.VERBOSE,
)

_comprehensive_pitch_class_name_regex = re.compile(
    '^{}$'.format(_comprehensive_pitch_class_name_regex_body),
    re.VERBOSE,
)

_comprehensive_pitch_name_regex = re.compile(
    '^{}$'.format(_comprehensive_pitch_name_regex_body),
    re.VERBOSE,
)

_pitch_class_name_regex = re.compile(
    '^{}$'.format(_pitch_class_name_regex_body),
    re.VERBOSE,
)

_pitch_class_octave_number_regex = re.compile(
    '^{}$'.format(_pitch_class_octave_number_regex_body),
    re.VERBOSE,
)

_pitch_name_regex = re.compile(
    '^{}$'.format(_pitch_name_regex_body),
    re.VERBOSE,
)

_range_string_regex = re.compile(
    '^{}$'.format(_range_string_regex_body),
    re.VERBOSE,
)

_interval_name_abbreviation_regex = re.compile(
    '^{}$'.format(_interval_name_abbreviation_regex_body),
    re.VERBOSE,
    )

del re



import collections
from abjad.system.AbjadObject import AbjadObject


class Lineage(AbjadObject, collections.Sequence):
    r'''
    Lineage of a component.

    ..  container:: example

        >>> score = abjad.Score()
        >>> staff = abjad.Staff(
        ...     r"""\new Voice = "Treble Voice" { c'4 }""",
        ...     name='Treble Staff',
        ...     )
        >>> score.append(staff)
        >>> bass = abjad.Staff(
        ...     r"""\new Voice = "Bass Voice" { b,4 }""",
        ...     name='Bass Staff',
        ...     )
        >>> score.append(bass)

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \context Staff = "Treble Staff"
                {
                    \context Voice = "Treble Voice"
                    {
                        c'4
                    }
                }
                \context Staff = "Bass Staff"
                {
                    \context Voice = "Bass Voice"
                    {
                        b,4
                    }
                }
            >>

        >>> for component in abjad.inspect(score).lineage():
        ...     component
        ...
        <Score<<2>>>
        <Staff-"Treble Staff"{1}>
        Voice("c'4", name='Treble Voice')
        Note("c'4")
        <Staff-"Bass Staff"{1}>
        Voice('b,4', name='Bass Voice')
        Note('b,4')

        >>> bass_voice = score['Bass Voice']
        >>> for component in abjad.inspect(bass_voice).lineage():
        ...     component
        ...
        <Score<<2>>>
        <Staff-"Bass Staff"{1}>
        Voice('b,4', name='Bass Voice')
        Note('b,4')

    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Selections'

    __slots__ = (
        '_component',
        '_components',
        )

    ### INITIALIZER ###

    def __init__(self, component=None):
        import abjad
        assert isinstance(component, (abjad.Component, type(None)))
        self._component = component
        components = []
        if component is not None:
            components.extend(
                reversed(
                abjad.inspect(component).parentage(include_self=False)))
            components.append(component)
            components.extend(
                abjad.inspect(component).descendants(include_self=False))
        self._components = components

    ### SPECIAL METHODS ###

    def __getitem__(self, argument):
        """
        Gets ``argument``.

        Returns component or tuple.
        """
        return self.components.__getitem__(argument)

    def __len__(self):
        """
        Gets length of lineage.

        Returns int.
        """
        return len(self._components)

    ### PUBLIC PROPERTIES ###

    @property
    def component(self):
        """
        The component from which the lineage was derived.
        """
        return self._component

    @property
    def components(self):
        """
        Gets components.

        Returns tuple.
        """
        return self._components



import abc
import logging
import os
import pickle
import ply  # type: ignore
import sys
from ply import yacc  # type: ignore
import traceback
from abjad.system.AbjadObject import AbjadObject


class Parser(AbjadObject):
    """
    Abstract base class for Abjad parsers.

    Rules objects for lexing and parsing must be defined by overriding the
    abstract properties ``lexer_rules_object`` and ``parser_rules_object``.

    For most parsers these properties should simply return ``self``.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_debug',
        '_lexer',
        '_logger',
        '_parser',
        )

    ### INITIALIZER ###

    def __init__(self, debug=False):

        self._debug = bool(debug)
        self._lexer = None
        self._parser = None

        if self.debug:
            logging.basicConfig(
                level=logging.DEBUG,
                filename=self.logger_path,
                filemode='w',
                format='%(filename)10s:%(lineno)8d:%(message)s'
                )
            self._logger = logging.getLogger()
        else:
            self._logger = yacc.NullLogger()

        self._lexer = ply.lex.lex(
            debug=self.debug,
            debuglog=self.logger,
            object=self.lexer_rules_object
            )

        if self.pickle_path and not os.path.exists(self.pickle_path):
            try:
                directory, _ = os.path.split(self.pickle_path)
                if not os.path.exists(directory):
                    os.makedirs(directory)
                string = pickle.dumps(None)
                with open(self.pickle_path, 'wb') as file_pointer:
                    file_pointer.write(string)
            except (IOError, OSError):
                traceback.print_exc()

        self._parser = ply.yacc.yacc(
            debug=self.debug,
            debuglog=self.logger,
            module=self.parser_rules_object,
            outputdir=self.output_path,
            picklefile=self.pickle_path,
            )

    ### SPECIAL METHODS ###

    def __call__(self, string):
        """
        Parse ``string`` and return result.
        """

        if hasattr(self, '_setup'):
            self._setup()

        if self.debug:
            result = self.parser.parse_debug(
                string,
                lexer=self.lexer,
                debug=self.logger,
                )
        else:
            result = self.parser.parse(string, lexer=self.lexer)
        if hasattr(self, '_cleanup'):
            result = self._cleanup(result)

        return result

    ### PUBLIC METHODS ###

    def tokenize(self, string):
        """
        Tokenize ``string`` and print results.
        """
        self.lexer.input(string)
        for token in self.lexer:
            print(token)

    ### PUBLIC PROPERTIES ###

    @property
    def debug(self):
        """
        Is true when parser runs in debugging mode.
        """
        return self._debug

    @property
    def lexer(self):
        """
        Gets parser's PLY Lexer instance.
        """
        return self._lexer

    @abc.abstractproperty
    def lexer_rules_object(self):
        """
        Gets object containing the parser's lexical rule definitions.
        """
        raise NotImplementedError

    @property
    def logger(self):
        """
        Gets parser's logger.
        """
        return self._logger

    @property
    def logger_path(self):
        """
        Gets parser's logfile output path.
        """
        if self.output_path is None:
            return None
        file_name = 'parselog_{}_{}.txt'.format(
            type(self).__name__,
            '-'.join(str(x) for x in sys.version_info),
            )
        return os.path.join(self.output_path, file_name)

    @property
    def output_path(self):
        """
        Gets output path for files associated with the parser.
        """
        from abjad import abjad_configuration
        configuration_directory = abjad_configuration.configuration_directory
        output_path = os.path.join(
            str(configuration_directory),
            'parsers',
            )
        if not os.path.isdir(output_path):
            try:
                os.makedirs(output_path)
            except (IOError, OSError):
                return None
        return output_path

    @property
    def parser(self):
        """
        Gets parser's PLY LRParser instance.
        """
        return self._parser

    @abc.abstractproperty
    def parser_rules_object(self):
        """
        Gets object containing the parser's syntactical rule definitions.
        """
        raise NotImplementedError

    @property
    def pickle_path(self):
        """
        Gets output path for the parser's pickled parsing tables.
        """
        if self.output_path is None:
            return None
        file_name = 'parse_tables_{}_{}.pkl'.format(
            type(self).__name__,
            '-'.join(str(x) for x in sys.version_info),
            )
        return os.path.join(self.output_path, file_name)



from .TimeRelation import TimeRelation


class TimespanTimespanTimeRelation(TimeRelation):
    r"""
    Timespan vs. timespan time relation.

    Score for examples:

    ..  container:: example

        >>> string = r"\times 2/3 { c'4 d'4 e'4 } \times 2/3 { f'4 g'4 a'4 }"
        >>> staff_1 = abjad.Staff(string)
        >>> staff_2 = abjad.Staff("c'2. d'4")
        >>> score = abjad.Score([staff_1, staff_2])

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        f'4
                        g'4
                        a'4
                    }
                }
                \new Staff
                {
                    c'2.
                    d'4
                }
            >>

        >>> last_tuplet = staff_1[-1]
        >>> long_note = staff_2[0]

        >>> abjad.show(score) # doctest: +SKIP

    ..  container:: example

        >>> abjad.timespans.timespan_2_happens_during_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    ..  container:: example

        >>> abjad.timespans.timespan_2_intersects_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        True

    ..  container:: example

        >>> abjad.timespans.timespan_2_is_congruent_to_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    ..  container:: example

        >>> abjad.timespans.timespan_2_overlaps_all_of_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    ..  container:: example

        >>> abjad.timespans.timespan_2_overlaps_start_of_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        True

    ..  container:: example

        >>> abjad.timespans.timespan_2_overlaps_stop_of_timespan_1(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    ..  container:: example

        >>> abjad.timespans.timespan_2_starts_after_timespan_1_starts(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    ..  container:: example

        >>> abjad.timespans.timespan_2_starts_after_timespan_1_stops(
        ...     timespan_1=last_tuplet,
        ...     timespan_2=long_note,
        ...     )
        False

    Timespan / timespan time relations are immutable.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Time relations'

    __slots__ = (
        '_timespan_1',
        '_timespan_2',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, inequality=None, timespan_1=None, timespan_2=None):
        TimeRelation.__init__(self, inequality=inequality)
        self._timespan_1 = timespan_1
        self._timespan_2 = timespan_2

    ### SPECIAL METHODS ###

    # TODO: hoist to TimeRelation
    def __call__(self, timespan_1=None, timespan_2=None):
        """
        Evaluate time relation.

        ..  container:: example

            Evaluate time relation without substitution:

            >>> timespan_1 = abjad.Timespan(5, 15)
            >>> timespan_2 = abjad.Timespan(10, 20)

            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

            >>> abjad.f(time_relation)
            abjad.timespans.TimespanTimespanTimeRelation(
                inequality=abjad.timespans.CompoundInequality(
                    [
                        abjad.TimespanInequality('timespan_1.start_offset <= timespan_2.start_offset'),
                        abjad.TimespanInequality('timespan_2.start_offset < timespan_1.stop_offset'),
                        ],
                    logical_operator='and',
                    ),
                timespan_1=abjad.Timespan(
                    start_offset=abjad.Offset(5, 1),
                    stop_offset=abjad.Offset(15, 1),
                    ),
                timespan_2=abjad.Timespan(
                    start_offset=abjad.Offset(10, 1),
                    stop_offset=abjad.Offset(20, 1),
                    ),
                )

            >>> time_relation()
            True

        ..  container:: example

            Substitute ``timespan_1`` during evaluation:

            >>> new_timespan_1 = abjad.Timespan(0, 10)

            >>> new_timespan_1
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(10, 1))

            >>> time_relation(timespan_1=new_timespan_1)
            False

        ..  container:: example

            Substitute ``timespan_2`` during evaluation:

            >>> new_timespan_2 = abjad.Timespan(2, 12)

            >>> new_timespan_2
            Timespan(start_offset=Offset(2, 1), stop_offset=Offset(12, 1))

            >>> time_relation(timespan_2=new_timespan_2)
            False

        ..  container:: example

            Substitute both ``timespan_1`` and ``timespan_2`` during evaluation:

            >>> time_relation(
            ...     timespan_1=new_timespan_1,
            ...     timespan_2=new_timespan_2,
            ...     )
            True

        Raise value error if either ``timespan_1`` or ``timespan_2`` is none.

        Otherwise return boolean.
        """
        import abjad
        timespan_1 = timespan_1 or self.timespan_1
        timespan_2 = timespan_2 or self.timespan_2
        if timespan_1 is None or timespan_2 is None:
            message = 'time relation is not fully loaded: {!r}.'
            raise ValueError(message.format(self))
        if not isinstance(timespan_1, abjad.Timespan):
            timespan_1 = abjad.Timespan()._get_timespan(timespan_1)
        if not isinstance(timespan_2, abjad.Timespan):
            timespan_2 = abjad.Timespan()._get_timespan(timespan_2)
        truth_value = self.inequality.evaluate(
            timespan_1.start_offset, timespan_1.stop_offset,
            timespan_2.start_offset, timespan_2.stop_offset)
        return truth_value

    def __eq__(self, argument):
        """
        Is true when ``argument`` equals time relation.

        ..  container:: example

            >>> timespan = abjad.Timespan(0, 10)
            >>> time_relation_1 = abjad.timespans.timespan_2_starts_during_timespan_1()
            >>> time_relation_2 = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan)

            >>> time_relation_1 == time_relation_1
            True
            >>> time_relation_1 == time_relation_2
            False
            >>> time_relation_2 == time_relation_2
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __hash__(self):
        """
        Hashes time relation.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super(TimespanTimespanTimeRelation, self).__hash__()

    ### PUBLIC METHODS ###

    def get_counttime_components(self, counttime_components):
        """
        Get ``counttime_components`` that satisfy ``time_relation``:

        ..  container:: example

            >>> notes = [abjad.Note(_ % 36, (1, 4)) for _ in range(200)]
            >>> voice = abjad.Voice(notes)
            >>> timespan_1 = abjad.Timespan(20, 22)
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1)

            >>> result = time_relation.get_counttime_components(voice[:])

            >>> for counttime_component in result:
            ...     counttime_component
            Note("af'4")
            Note("a'4")
            Note("bf'4")
            Note("b'4")
            Note("c''4")
            Note("cs''4")
            Note("d''4")
            Note("ef''4")

            >>> abjad.inspect(result).timespan()
            Timespan(start_offset=Offset(20, 1), stop_offset=Offset(22, 1))

        ``counttime_components`` must belong to a single voice.

        ``counttime_components`` must be time-contiguous.

        The call shown here takes 78355 function calls under r9686.

        Returns selection.
        """
        import abjad
        # check input
        assert isinstance(counttime_components, (
            list, abjad.Selection)), repr(counttime_components)
        assert self.timespan_1 is not None
        # iterate counttime components
        result = []
        for counttime_component in counttime_components:
            if self(timespan_2=abjad.inspect(counttime_component).timespan()):
                result.append(counttime_component)
        # return result
        return abjad.select(result)

    def get_offset_indices(
        self,
        timespan_2_start_offsets,
        timespan_2_stop_offsets,
        ):
        """
        Get offset indices that satisfy time relation:

        ..  container:: example

            >>> staff = abjad.Staff("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
            >>> start_offsets = [
            ...     abjad.inspect(note).timespan().start_offset
            ...     for note in staff
            ...     ]
            >>> stop_offsets = [
            ...     abjad.inspect(note).timespan().stop_offset
            ...     for note in staff
            ...     ]

        ..  container:: example

            Notes equal to ``staff[0:2]`` start during timespan ``[0, 3/16)``:

            >>> timespan_1 = abjad.Timespan((0), (3, 16))
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     )
            >>> time_relation.get_offset_indices(start_offsets, stop_offsets)
            (0, 2)

        ..  container:: example

            Notes equal to ``staff[2:8]`` start after timespan ``[0, 3/16)``
            stops:

            >>> timespan_1 = abjad.Timespan((0), (3, 16))
            >>> time_relation =  abjad.timespans.timespan_2_starts_after_timespan_1_stops(
            ...     timespan_1=timespan_1,
            ...     )
            >>> time_relation.get_offset_indices(start_offsets, stop_offsets)
            (2, 8)

        Returns nonnegative integer pair.
        """
        result = self.inequality.get_offset_indices(
            self.timespan_1,
            timespan_2_start_offsets,
            timespan_2_stop_offsets,
            )
        if not result:
            return []
        elif len(result) == 1:
            timespan = result[0]
            start_index = int(timespan.start_offset)
            stop_index = int(timespan.stop_offset)
            return start_index, stop_index
        elif 0 < len(result):
            message = 'inequality evaluates to disjunct range: {!r}.'
            message = message.format(result)
            raise Exception(message)

    ### PUBLIC PROPERTIES ###

    @property
    def is_fully_loaded(self):
        """
        Is true when ``timespan_1`` and ``timespan_2`` are both not none.

        ..  container:: example

            >>> timespan_1 = abjad.Timespan(0, 10)
            >>> timespan_2 = abjad.Timespan(5, 15)
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

            >>> time_relation.is_fully_loaded
            True

        Returns true or false.
        """
        return self.timespan_1 is not None and self.timespan_2 is not None

    @property
    def is_fully_unloaded(self):
        """
        Is true when ``timespan_1`` and ``timespan_2`` are both none.

        ..  container:: example

            >>> timespan_1 = abjad.Timespan(0, 10)
            >>> timespan_2 = abjad.Timespan(5, 15)
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

            >>> time_relation.is_fully_unloaded
            False

        Returns true or false.
        """
        return self.timespan_1 is None and self.timespan_2 is None

    @property
    def timespan_1(self):
        """
        Time relation timespan ``1``:

        ..  container:: example

            >>> timespan_1 = abjad.Timespan(0, 10)
            >>> timespan_2 = abjad.Timespan(5, 15)
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

            >>> time_relation.timespan_1
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(10, 1))

        Returns timespan.
        """
        return self._timespan_1

    @property
    def timespan_2(self):
        """
        Time relation timespan ``2``:

        ..  container:: example

            >>> timespan_1 = abjad.Timespan(0, 10)
            >>> timespan_2 = abjad.Timespan(5, 15)
            >>> time_relation = abjad.timespans.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

            >>> time_relation.timespan_2
            Timespan(start_offset=Offset(5, 1), stop_offset=Offset(15, 1))

        Returns timespan.
        """
        return self._timespan_2



import typing
from abjad.system.AbjadObject import AbjadObject
from abjad.utilities.String import String
from abjad.system.FormatSpecification import FormatSpecification
from .Part import Part


class Section(AbjadObject):
    """
    Section.

    ..  container:: example

        >>> abjad.Section(
        ...     abbreviation='VN-1',
        ...     count=18,
        ...     instrument='Violin',
        ...     name='FirstViolin',
        ...     )
        Section(abbreviation='VN-1', count=18, instrument='Violin', name='FirstViolin')

        >>> abjad.Section(
        ...     abbreviation='VN-2',
        ...     count=18,
        ...     instrument='Violin',
        ...     name='SecondViolin',
        ...     )
        Section(abbreviation='VN-2', count=18, instrument='Violin', name='SecondViolin')

        >>> abjad.Section(
        ...     abbreviation='VA',
        ...     count=18,
        ...     name='Viola',
        ...     )
        Section(abbreviation='VA', count=18, instrument='Viola', name='Viola')

        >>> abjad.Section(
        ...     abbreviation='VC',
        ...     count=14,
        ...     name='Cello',
        ...     )
        Section(abbreviation='VC', count=14, instrument='Cello', name='Cello')

        >>> abjad.Section(
        ...     abbreviation='CB',
        ...     count=6,
        ...     name='Contrabass',
        ...     )
        Section(abbreviation='CB', count=6, instrument='Contrabass', name='Contrabass')

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_abbreviation',
        '_count',
        '_instrument',
        '_name',
        '_parts',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        abbreviation: str = None,
        count: int = 1,
        instrument: str = None,
        name: str = None,
        ) -> None:
        if abbreviation is not None:
            assert isinstance(abbreviation, str), repr(abbreviation)
        self._abbreviation = abbreviation
        if not isinstance(count, int):
            raise Exception(f'Count must be integer (not {count!r}).')
        if not 1 <= count:
            raise Exception(f'Count must be positive (not {count!r}).')
        self._count = count
        if instrument is not None:
            assert isinstance(instrument, str), repr(instrument)
        else:
            instrument = name
        self._instrument = instrument
        if name is not None:
            assert isinstance(name, str), repr(name)
        self._name = name
        parts = []
        if self.count is None:
            part = Part(self.name)
            parts.append(part)
        else:
            if 1 < len(str(self.count)):
                zfill: typing.Optional[int] = len(str(self.count))
            else:
                zfill = None
            for member in range(1, self.count + 1):
                part = Part(
                    member=member,
                    instrument=self.instrument,
                    section=self.name,
                    section_abbreviation=self.abbreviation,
                    zfill=zfill,
                    )
                parts.append(part)
        self._parts = parts

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a section with the same name,
        abbreviation and count as this section.

        ..  container:: example

            >>> section_1 = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )
            >>> section_2 = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )
            >>> section_3 = abjad.Section(
            ...     abbreviation='VN-2',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='SecondViolin',
            ...     )

            >>> section_1 == section_1
            True
            >>> section_1 == section_2
            True
            >>> section_1 == section_3
            False

            >>> section_2 == section_1
            True
            >>> section_2 == section_2
            True
            >>> section_2 == section_3
            False

            >>> section_3 == section_1
            False
            >>> section_3 == section_2
            False
            >>> section_3 == section_3
            True

        """
        if (isinstance(argument, type(self)) and
            argument.name == self.name and
            argument.abbreviation == self.abbreviation and
            argument.count == self.count):
            return True
        return False

    def __hash__(self):
        """
        Hashes section.
        """
        return super().__hash__()

    ### PUBLIC PROPERTIES ###

    @property
    def abbreviation(self) -> typing.Optional[str]:
        """
        Gets abbreviation.

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )

            >>> section.abbreviation
            'VN-1'

        """
        return self._abbreviation

    @property
    def count(self) -> typing.Optional[int]:
        """
        Gets section count.

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )

            >>> section.count
            18

        """
        return self._count

    @property
    def instrument(self) -> typing.Optional[str]:
        """
        Gets section instrument.

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )

            >>> section.instrument
            'Violin'

        """
        return self._instrument

    @property
    def name(self) -> typing.Optional[str]:
        """
        Gets section name.

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )

            >>> section.name
            'FirstViolin'

        """
        return self._name

    @property
    def parts(self) -> typing.List[Part]:
        """
        Gets parts.

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )

            >>> for part in section.parts:
            ...     part
            ...
            Part(instrument='Violin', member=1, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=2, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=3, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=4, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=5, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=6, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=7, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=8, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=9, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=10, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=11, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=12, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=13, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=14, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=15, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=16, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=17, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=18, section='FirstViolin', section_abbreviation='VN-1', zfill=2)

        """
        return list(self._parts)
